module InfiniteMap where

import Prelude hiding (null)

import Data.Map as Map

 -- An infinite map library.

data InfiniteMap k a = InfMap (Maybe a, Map.Map k a)
 deriving (Ord)

instance (Eq a, Ord k, Show k, Show a) => Show (InfiniteMap k a) where
 show (InfMap (Nothing,m)) = (show m)
 show (InfMap (Just y,m)) | null m = "{_:=" ++ (show y) ++ "}"
                          | otherwise = "{_:=" ++ (show y) ++ "}" ++ sm'
                             where m' = Map.filter (\ y' -> not(y' == y)) m
                                   sm' = if null m' then "" else show m'

(!)    :: Ord k => InfiniteMap k a -> k -> a

constant :: a -> InfiniteMap k a
empty    :: InfiniteMap k a


insert    :: Ord k => k -> a -> InfiniteMap k a ->  InfiniteMap k a
insertAll :: Ord k => [(k,a)] -> InfiniteMap k a -> InfiniteMap k a
but :: Ord k => InfiniteMap k a -> [(k,a)] -> InfiniteMap k a

InfMap(Just y,m)  ! k = Map.findWithDefault y k m
InfMap(Nothing,m) ! k = m Map.! k

lookup :: (Ord k,Monad m) => k -> InfiniteMap k a -> m a
lookup k (InfMap(Just y,m)) = return (findWithDefault y k m)
lookup k (InfMap(Nothing,m)) = Map.lookup k m

empty = InfMap(Nothing,Map.empty)

constant y = InfMap(Just y,Map.empty)

insert k a (InfMap(def,m)) = InfMap(def,Map.insert k a m)

insertAll [] im = im
insertAll ((k,a):bindings) im = InfiniteMap.insert k a (insertAll bindings im)



isSubmapOfBy :: Ord k => (a -> b -> Bool) -> InfiniteMap k a -> InfiniteMap k b -> Bool
isSubmapOfBy (<:) (InfMap(def1,m1)) (InfMap(def2,m2)) = 
 case (def1,def2) of 
  (Just _,Nothing) -> False
  (Just y1,Just y2) -> y1 <: y2 && isSubmap
  _ -> isSubmap
  where
   isSubmap = Map.foldWithKey (\ k a tally -> 
                tally && case (Map.lookup k m2) of
                          Just b -> a <: b
                          Nothing -> False) True m1



isSubmapOf :: (Ord k, Eq a) => InfiniteMap k a -> InfiniteMap k a -> Bool
isSubmapOf m1 m2 = InfiniteMap.isSubmapOfBy (==) m1 m2


lift :: (Ord k) => (a -> b -> c) -> InfiniteMap k a
                                 -> InfiniteMap k b
                                 -> InfiniteMap k c
lift f (InfMap(Just ya,ma)) (InfMap(Just yb,mb)) =
 let map' = Map.foldWithKey (\ k a newMap -> 
            Map.insert k (f a (Map.findWithDefault yb k mb)) newMap) 
            (Map.empty) ma
     map'' = Map.foldWithKey (\ k b newMap ->
             Map.insert k (f (Map.findWithDefault ya k ma) b) newMap)
             map' mb
 in  InfMap(Just (f ya yb),map'')
  

instance (Ord k, Eq a) => Eq (InfiniteMap k a) where
 im1 == im2 = (InfiniteMap.isSubmapOf im1 im2) && (InfiniteMap.isSubmapOf im2 im1)


map f (InfMap(Just y,m)) = (InfMap(Just (f y),Map.map f m))
map f (InfMap(Nothing,m)) = (InfMap(Nothing,Map.map f m))


filterWithKey :: a -> (k -> a -> Bool) -> InfiniteMap k a -> InfiniteMap k a
filterWithKey dead p (InfMap(def,m)) = 
 InfMap(def,Map.mapWithKey(\ k a -> if p k a then a else dead) m)


im `but` xys = insertAll xys im

foldWithKey :: (Maybe a -> b) -> (k -> a -> b -> b) -> InfiniteMap k a -> b
foldWithKey init f (InfMap(my,m)) = Map.foldWithKey f (init my) m

fold :: (Maybe a -> b) -> (a -> b -> b) -> InfiniteMap k a -> b
fold init f (InfMap(my,m)) = Map.fold f (init my) m

toList :: (Ord k, Eq a) => InfiniteMap k a -> [(Maybe k,a)]
toList (InfMap(Just y,m)) = 
 if valid m
 then  (Nothing,y):[ (Just k,a) | (k,a) <- Map.toList m, not(a == y) ]
 else error "map is in invalid state"
toList (InfMap(Nothing,m)) = [ (Just k,a) | (k,a) <- Map.toList m ]