module Might where

import Data.Set as Set
import Control.Monad.State


 -- Syntactic sugar.

(==>) a b = (a,b)



 -- Debugging aids.

blank = error "blank" 
todo s = error ("todo: " ++ s)
zipFail errmsg l1 l2 = if length l1 == length l2 
                       then zip l1 l2
                       else error $ "zipFail: " ++ errmsg


except :: Eq a => (a -> b) -> (a,b) -> (a -> b)
except f (x,y) = \ x' -> if x == x' then y else f x'

fix init f = if f init == init
             then init 
             else f (f init)


 -- List manipulation.

foldOp f init [] =  init
foldOp f init (hd:tl) = foldOp f (f init hd) tl

appendReverse [] ans = ans
appendReverse (hd:tl) ans = appendReverse tl (hd:ans)

flattenListOfSets :: [Set.Set a] -> [a]
flattenListOfSets lst = 
 foldl (\ tally set -> (Set.toList set) ++ tally) [] lst


 -- Foldable.

class Foldable m where 
 xfold :: (a -> b -> b) -> b -> m a -> b


instance Foldable [] where
 xfold = foldr

instance Foldable Set.Set where
 xfold = Set.fold
 

 -- Monad utilities.

doFoldAction :: (Monad m, Foldable f) => f a -> (a -> m ()) -> m ()
doFoldAction foldable action = 
 xfold (\ a mon -> mon >> action a) (return ()) foldable


 -- State monad extensions.

retrieve property = 
 do state <- get
    return $ property state

perform f =
 do state <- get
    put $ f state
