module AbstractFrameStrings where

import Might

import InfiniteMap as IMap
import Data.Set as Set
import Lattice

import DS
import CPS


 -- Frame strings.

data Delta = E | D | U | DDP | UUP | UPDP
 deriving(Eq,Ord,Show)


instance Lattice Delta where
 top = error "top Delta"
 bot = error "bot Delta"
 (<:) = (==)


type PowDelta = Set.Set Delta

topPowDelta = Set.fromList[ E, D, U, DDP, UUP, UPDP ]

type F = InfiniteMap LAB PowDelta

topF = IMap.constant topPowDelta

abscat :: Delta -> Delta -> PowDelta
abscat E   a    = Set.fromList[a]
abscat a   E    = Set.fromList[a]

abscat D   D    = Set.fromList[DDP]
abscat D   DDP  = Set.fromList[DDP]
abscat D   U    = Set.fromList[E]
abscat D   UUP  = Set.fromList[U, UUP]
abscat D   UPDP = Set.fromList[DDP, UPDP]
		      
abscat U   D    = Set.fromList[E, UPDP]
abscat U   DDP  = Set.fromList[D, DDP, UPDP]
abscat U   U    = Set.fromList[UUP]
abscat U   UUP  = Set.fromList[UUP]
abscat U   UPDP = Set.fromList[UPDP]

abscat DDP D    = Set.fromList[DDP]
abscat DDP DDP  = Set.fromList[DDP]
abscat DDP U    = Set.fromList[D, DDP]
abscat DDP UUP  = Set.fromList[E, D, U, DDP, UUP]
abscat DDP UPDP = Set.fromList[D, DDP, UPDP]		      

abscat UUP D    = Set.fromList[U, UUP]
abscat UUP DDP  = Set.fromList[E, D, U, DDP, UUP, UPDP]
abscat UUP U    = Set.fromList[UUP]
abscat UUP UUP  = Set.fromList[UUP]
abscat UUP UPDP = Set.fromList[UPDP]

abscat UPDP D    = Set.fromList[UPDP]
abscat UPDP DDP  = Set.fromList[UPDP]
abscat UPDP U    = Set.fromList[U, UUP, UPDP]
abscat UPDP UUP  = Set.fromList[U, UUP, UPDP]
abscat UPDP UPDP = Set.fromList[E, D, U, DDP, UUP, UPDP]

cat l r = 
 Set.unions[ a1 `abscat` a2 | a1 <- Set.toList l, a2 <- Set.toList r ]

catF :: F -> F -> F
catF f g = if (f == bot) 
           then bot
           else IMap.lift cat f g

class Invertible a where
 inv :: a -> a

instance Invertible Delta where
 inv E = E
 inv D = U
 inv U = D
 inv UUP = DDP
 inv DDP = UUP
 inv UPDP = UPDP

instance (Ord a, Invertible a) => Invertible (Set.Set a) where
 inv set = Set.map inv set

instance (Ord k, Invertible a) => Invertible (IMap.InfiniteMap k a) where
 inv im = IMap.map inv im

emptyString :: F
emptyString = constant (singleton E)

downFor :: LAB -> F
downFor l = emptyString `but` [l ==> singleton D]

eqv_cn :: F -> F -> Bool
f `eqv_cn` g | f == bot || (inv g) == bot = False
             | otherwise            = pureCont(f `catF` (inv g))
 where pureCont p = IMap.foldWithKey (\ def -> case def of
                      Nothing -> error "empty frame string?"
                      Just pd -> pd == singleton E)
                     (\ lab pd tally -> 
                         if not(tally)
                         then False
                         else case lab of 
                               Cn _ -> True
                               Fn _ -> pd == singleton E)
                     p

eqv :: F -> F -> Bool                     
f `eqv` g | f == bot || (inv g) == bot = False
          | otherwise            = (f `catF` (inv g)) == emptyString

