 -- A small, macro-less Scheme.
module App.ThScheme.Main where

import qualified Data.Maybe as Maybe
import qualified Data.Map as Map
import qualified Data.Set as Set

import qualified System (getArgs)
import qualified Control.Monad.State as St

import qualified Parser.SExpParser as SXP

import qualified Compiler.Symbol as Sym
import qualified Compiler.SExpression as SX
import qualified Compiler.RxRS as DS

import qualified Compiler.GCFA as GCFA

import qualified Might as Might

dsAnt = DS.botAnn


parseFormal :: SX.LexicalSExp -> DS.Term ()
parseFormal (SX.Symbol sym annt) = (DS.Var sym dsAnt)


bindsPrim term@(DS.Var sym annt) = 
 case parseExp (SX.Symbol sym SX.SourceUndefined) of
  (DS.Prim{}) -> True
  _ -> False
bindsPrim term@(DS.Forms ordered labels labeled rest annt) = 
 any bindsPrim ((Maybe.maybeToList rest) ++ (ordered ++ labeled))




parseExp :: SX.LexicalSExp -> DS.Term ()
parseExp (SX.Nil annt) = DS.Lit DS.LitNil dsAnt
parseExp (SX.Int int annt) = DS.Lit (DS.LitInt int) dsAnt
parseExp (SX.Char char annt) = DS.Lit (DS.LitChar char) dsAnt
parseExp (SX.Bool bool annt) = DS.Lit (DS.LitBool bool) dsAnt
parseExp (SX.String string annt) = DS.Lit (DS.LitString string) dsAnt 
parseExp (SX.Symbol sym annt) = 
 case Sym.toString sym of

   -- Arithmetic.
  "+"  -> (DS.Prim (DS.SUM) dsAnt)
  "-"  -> (DS.Prim (DS.DIFF) dsAnt)
  "*"  -> (DS.Prim (DS.PROD) dsAnt)
  "/"  -> (DS.Prim (DS.DIV) dsAnt)

  "remainder" -> (DS.Prim (DS.REMAINDER) dsAnt)
  "quotient" -> (DS.Prim (DS.QUOTIENT) dsAnt)
  
  "exact->inexact" -> (DS.Prim (DS.EXACT2INEXACT) dsAnt)

  "max" -> (DS.Prim (DS.MAX) dsAnt)
  "min" -> (DS.Prim (DS.MIN) dsAnt)
  "expt" -> (DS.Prim (DS.EXPT) dsAnt)

   -- Strings
  "number->string" -> (DS.Prim (DS.NUMBER2STRING) dsAnt)
  "string-append" -> (DS.Prim (DS.STRING_APPEND) dsAnt)

   -- Arithmetic relations.
  "="  -> (DS.Prim (DS.EQNUM) dsAnt)
  "<"  -> (DS.Prim (DS.LT) dsAnt)
  "<=" -> (DS.Prim (DS.LEQ) dsAnt)
  ">"  -> (DS.Prim (DS.GT) dsAnt)
  ">=" -> (DS.Prim (DS.GEQ) dsAnt)
  "number?" -> (DS.Prim (DS.NUMBERP) dsAnt)
  "zero?" -> (DS.Prim (DS.ZEROP) dsAnt)


   -- Functions.
  "apply" -> (DS.Prim (DS.APPLY) dsAnt)

   -- Vectors.
  "vector-length" -> (DS.Prim (DS.VECTOR_LENGTH) dsAnt)
  "make-vector" -> (DS.Prim (DS.MAKE_VECTOR) dsAnt)
  "vector" -> (DS.Prim (DS.VECTOR) dsAnt)
  "vector-set!" -> (DS.Prim (DS.VECTORSET) dsAnt)
  "vector-ref" -> (DS.Prim (DS.VECTORREF) dsAnt)
 
   -- Lists.
  "cons"     -> (DS.Prim (DS.CONS) dsAnt)
  "car"      -> (DS.Prim (DS.CAR) dsAnt)
  "cdr"      -> (DS.Prim (DS.CDR) dsAnt)
  "set-car!" -> (DS.Prim (DS.SETCAR) dsAnt)
  "set-cdr!" -> (DS.Prim (DS.SETCDR) dsAnt)
  "list"     -> (DS.Prim (DS.LIST) dsAnt)
  "append"   -> (DS.Prim (DS.APPEND) dsAnt)
--   "memv"     -> (DS.Prim (DS.MEMV) dsAnt)
--   "memq"     -> (DS.Prim (DS.MEMQ) dsAnt)
  "null?"    -> (DS.Prim (DS.NULLP) dsAnt)
  "pair?"    -> (DS.Prim (DS.PAIRP) dsAnt)

   -- Equality.
  "eq?"      -> (DS.Prim (DS.EQP) dsAnt)
  "eqv?"     -> (DS.Prim (DS.EQVP) dsAnt)
  "equal?"   -> (DS.Prim (DS.EQUALP) dsAnt)

   -- Logic. 
  "not"      -> (DS.Prim (DS.NOT) dsAnt)

   -- NOTE: These are desugared; they are not handled like normal primitives.
   -- TODO: Lift these to term-level.
  "or"       -> (DS.Prim (DS.OR) dsAnt)
  "and"      -> (DS.Prim (DS.AND) dsAnt)

   -- I/O.
  "write"    -> (DS.Prim (DS.WRITE) dsAnt)
  "display"  -> (DS.Prim (DS.DISPLAY) dsAnt)
  "newline"  -> (DS.Prim (DS.NEWLINE) dsAnt)
 
   -- Errors.
  "error"    -> (DS.Prim (DS.ERROR) dsAnt)

   -- Control.
  "call/cc"  -> (DS.Prim (DS.CALLCC) dsAnt)
  "halt"     -> (DS.Prim (DS.HALT) dsAnt)

   -- Uninitialized values.
  ".unspecified" -> (DS.Prim (DS.UNSPECIFIED) dsAnt)

   -- Abstract-only primitives.
  "abstrace" -> (DS.Prim (DS.ABSTRACE) dsAnt)

  _ -> DS.Ref sym dsAnt

parseExp (SX.Cons f@(SX.Symbol{SX.symbol=Sym.Sym{Sym.toString=s}}) tl an)=
 case s of 

  -- (define <var> <exp>)
  -- (define (<var> <var>*) <exp>*)
  -- (define (<var> <var>* . <var>) <exp>*)
  "define" -> 
    let
     sxlhs = SX.car tl  
     sxexp = (SX.car . SX.cdr) tl
     sxbodyseq = SX.cdr tl
    in 
     if SX.isSymbol sxlhs
     then DS.Def (DS.Forms [parseFormal sxlhs] [] [] Nothing dsAnt)
                  (parseExp sxexp)
                  -- (DS.Args [parseExp sxexp] [] [] Nothing dsAnt)
                  dsAnt
     else 
      let
       sxvar = SX.car sxlhs
       sxforms = SX.cdr sxlhs

       (dsvars,dsrest) = SX.mapToImproperList parseFormal sxforms 

       var = DS.Forms [parseFormal sxvar] [] [] Nothing dsAnt
       forms = DS.Forms dsvars [] [] dsrest dsAnt
       dsbodyseq = DS.Seq (SX.mapToList parseExp sxbodyseq) dsAnt
      in DS.Def var (DS.Lam forms dsbodyseq dsAnt) dsAnt



  -- (do ((<var> <init> <step>)*) 
  --     (<test> <exp>*)
  --     <cmd>*)
  "do" ->
   let
     sxclauses = SX.car tl
     sxtermclause = (SX.car . SX.cdr) tl
     sxcommands = (SX.cdr . SX.cdr) tl

     parseDoClause (SX.Cons sxvar (SX.Cons sxinit (SX.Cons sxstep SX.Nil{} _) _) _)
      = (parseFormal sxvar, parseExp sxinit, parseExp sxstep)

     parseDoClause (SX.Cons sxvar (SX.Cons sxinit SX.Nil{} _) _)
      = (parseFormal sxvar, parseExp sxinit, parseExp sxvar)

     varinitsteps = SX.mapToList parseDoClause sxclauses

     (vars,inits,steps) = unzip3 varinitsteps

     dsterm = parseExp (SX.car sxtermclause)
     dsbody = DS.Seq (SX.mapToList parseExp (SX.cdr sxtermclause)) dsAnt
     
     dscmds = DS.Seq (SX.mapToList parseExp sxcommands) dsAnt

    in DS.Do vars inits steps dsterm dsbody dscmds dsAnt


  -- (lambda <var> <body>*)
  -- (lambda (<var>*) <body>*)
  -- (lambda (<var>* . <var>) <body>*)
  "lambda" ->
    let 
     sxvars = SX.car tl
     sxbodyseq = SX.cdr tl

     (dsvars,dsrest) = SX.mapToImproperList parseFormal sxvars
     dsbodyseq = DS.Seq (SX.mapToList parseExp sxbodyseq) dsAnt
    in DS.Lam (DS.Forms dsvars [] [] dsrest dsAnt) dsbodyseq dsAnt


  -- (begin <exp>*)
  "begin" ->  
   let
    sxexps = tl
    exps = SX.mapToList parseExp sxexps
    in DS.Seq exps dsAnt


  -- (set! <var> <exp>)
  "set!" ->
   let
    sxvar = SX.car tl
    sxexp = (SX.car . SX.cdr) tl

    dsvar = parseExp sxvar
    dsexp = parseExp sxexp

    in DS.Set dsvar dsexp (DS.Void dsAnt) dsAnt


  -- (let* ((<var> <exp>)*) <body>*)
  "let*" ->
    let 
     varexpsxs = SX.mapToList SX.toList (SX.car tl)
     sxbodyseq = SX.cdr tl

     dsbodyseq = DS.Seq (SX.mapToList parseExp sxbodyseq) dsAnt

     toNestedLets [] = dsbodyseq
     toNestedLets (([v,e]):tl) = 
      DS.Let [DS.Forms [parseFormal v] [] [] Nothing dsAnt] [(parseExp e)]
             (toNestedLets tl) dsAnt

    in toNestedLets varexpsxs


  -- (let ((<var> <exp>)*) <body>*)
  -- (let <loop-var> ((<var> <exp>)*) <body>*)
  "let" ->  case tl of
     (SX.Cons loopvar@(SX.Symbol{SX.symbol=loopsym}) cdr _) -> 
      let
       sxvarexps = SX.car cdr
       sxloopbody = SX.cdr cdr

       convVarInit (SX.Cons sxvar (SX.Cons sxexp (SX.Nil{}) _) _) =
        (parseFormal sxvar,parseExp sxexp)

       varinits = SX.mapToList convVarInit sxvarexps

       (vars,inits) = unzip varinits

       dsloopvar = parseFormal loopvar
       dsloopref = parseExp loopvar

       dsloopbody = DS.Seq (SX.mapToList parseExp sxloopbody) dsAnt

       dslam = DS.Lam (DS.Forms vars [] [] Nothing dsAnt) dsloopbody dsAnt

       in
        DS.LetRec
         [DS.formalized [dsloopvar]]
         [dslam]
         (DS.App dsloopref (DS.Args inits [] [] Nothing dsAnt) dsAnt)
         dsAnt

     _ -> 
      let
       varexpsxs = SX.mapToList SX.toList (SX.car tl)
       sxbodyseq = SX.cdr tl

       parseVarExp [v,e] = (DS.Forms [parseFormal v] [] [] Nothing dsAnt,
  			  parseExp e)

       dsbodyseq = DS.Seq (SX.mapToList parseExp sxbodyseq) dsAnt
 
       (vars,exps) = unzip (map parseVarExp varexpsxs)

       in DS.Let vars exps dsbodyseq dsAnt


  -- (letrec ((<var> <exp>)*) <body>*)
  "letrec" ->
    let 
     varexpsxs = SX.mapToList SX.toList (SX.car tl)
     sxbodyseq = SX.cdr tl

     parseVarExp [v,e] = (DS.Forms [parseFormal v] [] [] Nothing dsAnt,
			  parseExp e)

     dsbodyseq = DS.Seq (SX.mapToList parseExp sxbodyseq) dsAnt
 
     (vars,exps) = unzip (map parseVarExp varexpsxs)
    in DS.LetRec vars exps dsbodyseq dsAnt


  -- (if <cond> <consequent> <alternate>)
  -- (if <cond> <consequent>)
  "if" -> 
   let
    sxcond = SX.car tl
    sxconsequent = (SX.car . SX.cdr) tl
    sxalternate  = (SX.car . SX.cdr . SX.cdr) tl

    cond = parseExp sxcond
    consequent = parseExp sxconsequent
    alternate = parseExp sxalternate

   in if SX.isCons $ (SX.cdr . SX.cdr) tl
      then DS.Cond [cond,DS.Lit (DS.LitBool True) dsAnt] 
                    [consequent,alternate] dsAnt
      else DS.Cond [cond] [consequent] dsAnt


  -- (cond (<cond> <consequent>*)* [(else <exp>)])
  "cond" ->
   let
    sxclauses = tl
    
    parseClause (SX.Cons (SX.Symbol{SX.symbol=Sym.Sym {Sym.toString="else"}}) action _) =
     -- TODO: Parse everything behind here into a Seq, too.
     (DS.Lit (DS.LitBool True) dsAnt, parseExp (SX.car action))
    parseClause (SX.Cons cond action _) =
     (parseExp cond, 
      DS.Seq (SX.mapToList parseExp action) dsAnt)

    condsthens = SX.mapToList parseClause sxclauses

    (conds,thens) = unzip condsthens
   in DS.Cond conds thens dsAnt


  -- (case <exp> ((<lit>*) <exp>*)* [(else <exp>)])
  "case" ->
   let
    sxval = SX.car tl
    sxclauses = SX.cdr tl

    parseClause (SX.Cons (SX.Symbol{SX.symbol=Sym.Sym {Sym.toString="else"}}) action _) =
     (Nothing, parseExp (SX.car action))
    parseClause (SX.Cons lits action _) =
     (Just $! SX.mapToList parseLit lits, 
      DS.Seq (SX.mapToList parseExp action) dsAnt)

    (litss,thens) = unzip (SX.mapToList parseClause sxclauses)

   in DS.Case (parseExp sxval) litss thens dsAnt

   
  -- (quote <lit>) 
  "quote" ->
   let
    sxlit = SX.car tl
   in parseLit sxlit 


  -- (try <exp> <handler>)
--   "try" ->
--    let
--     sxexp = SX.kar tl
--     sxhandler = (SX.kar . SX.cdr) tl
    
--     exp = parseExp sxexp
--     handler = parseExp sxhandler
--    in DS.Try exp handler dsAnt


  -- (raise <exn>)
--   "raise" ->
--    let

--     sxexn = SX.kar tl
    
--     exn = parseExp sxexn
--    in DS.Raise exn dsAnt
  
  _ -> DS.App (parseExp f) 
       (DS.Args (SX.mapToList parseExp tl) [] [] Nothing dsAnt)
       dsAnt
   
     
parseExp (SX.Cons f args annt) =
 DS.App (parseExp f)
        (DS.Args (SX.mapToList parseExp args) [] [] Nothing dsAnt) dsAnt

-- parseExp (SX.Nil annt) =
--  DS.Lit (DS.LitNil) dsAnt

parseExp other = 
 error $! "unknown S-Exp: " ++ (show other)



 -- parseLit: Parse an S-Expression into its literal self.
parseLit (SX.Nil{}) =
 DS.Lit (DS.LitNil) dsAnt

parseLit (SX.Int{SX.int=int}) = 
 DS.Lit (DS.LitInt int) dsAnt

parseLit (SX.Symbol{SX.symbol=sym}) = 
 DS.Lit (DS.LitSymbol sym) dsAnt

parseLit (SX.Bool{SX.bool=truth}) = 
 DS.Lit (DS.LitBool truth) dsAnt

parseLit (SX.String{SX.string=str}) = 
 DS.Lit (DS.LitString str) dsAnt

parseLit (SX.Char{SX.character=c}) = 
 DS.Lit (DS.LitChar c) dsAnt

parseLit sx@(SX.Cons car cdr _) = 
 DS.App (DS.Prim DS.CONS dsAnt) (DS.Args [carLit,cdrLit] [] [] Nothing dsAnt) dsAnt
 where 
  carLit = parseLit car
  cdrLit = parseLit cdr
 
parseLit other = error $ "parseLit? " ++ (show other)






 -- Read in a file.  Analyze it.
test :: IO ()
test =
 do
 args <- System.getArgs

 arg0 <- return $! args !! 0

 contents <- readFile arg0

 stdlib <- readFile "stdlib.scm"

 -- sxs <- return $! SX.listFromStringWithName arg0 contents
 sxs <- return $! SXP.listFromStringWithName arg0 (stdlib ++ contents)
 
 (sxs,st) <- return $! St.runState (SX.canonicalizeMany sxs) Sym.empty

 putStrLn ""
 putStrLn "canonicalized: "
 putStrLn (show sxs)
 putStrLn ""

 putStrLn ""
 putStrLn "parsed: "
 seqs <- return $! map parseExp sxs
 pr <- return $! DS.Seq seqs dsAnt
 putStrLn (show pr)

 putStrLn ""
 putStrLn "def2letrec-converted:"
 pr <- return $! DS.seqConvLetRec pr
 putStrLn (show pr)

 writeFile "debug/nodef.out.scm" (show pr)

 putStrLn ""
 putStrLn "free variable analysis: "
 pr <- return $! DS.annotateAllFree pr
 putStrLn (show (map fst (Map.toList (DS.free pr))))



 putStrLn ""
 putStrLn "let-wrapped:"
 pr <- return $! DS.letWrap pr
 putStrLn (show pr)

 writeFile "debug/let-wrap.out.scm" (show pr)


 putStrLn ""
 putStrLn "alphatized: "
 (pr,st) <- return $! St.runState (DS.alphatized pr) st
 putStrLn (show pr)




 putStrLn ""
 putStrLn "and/or-converted; do-converted: "
 (pr,st) <- return $! St.runState (DS.andOrConv pr) st
 (pr,st) <- return $! St.runState (DS.doConv pr) st
 putStrLn (show pr)

 writeFile "debug/andor.out.scm" (show pr)

 putStrLn ""
 putStrLn "free variable analysis: "
 pr <- return $! DS.annotateAllFree pr
 putStrLn (show (map fst (Map.toList (DS.free pr))))



 putStrLn ""
 putStrLn "letrec/let-converted: "
 (pr,st) <- return $! St.runState (DS.letrecConv pr) st
 putStrLn (show pr)

 writeFile "debug/letrec-conv.out.scm" (show pr)

 putStrLn ""
 putStrLn "free variable analysis: "
 pr <- return $! DS.annotateAllFree pr
 putStrLn (show (map fst (Map.toList (DS.free pr))))



 putStrLn ""
 putStrLn "syntactically dead code removal: "
 pr <- return $! DS.synDeadRemove pr
 putStrLn (show pr)

 writeFile "debug/no-dead.out.scm" (show pr)

 putStrLn ""
 putStrLn "free variable analysis: "
 pr <- return $! DS.annotateAllFree pr
 putStrLn (show (map fst (Map.toList (DS.free pr))))




 putStrLn ""
 putStrLn "case->cond converted: "
 (pr,st) <- return $! St.runState (DS.case2cond pr) st
 putStrLn (show pr)

 writeFile "debug/case2cond.out.scm" (show pr)

 putStrLn ""
 putStrLn "free variable analysis: "
 pr <- return $! DS.annotateAllFree pr
 putStrLn (show (map fst (Map.toList (DS.free pr))))

 putStrLn ""
 putStrLn "a-normalized: "
 (pr,st) <- return $! St.runState (DS.anormalized pr) st
 putStrLn (show pr)

 writeFile "debug/anorm.out.scm" (show pr)

 putStrLn ""
 putStrLn "free variable analysis: "
 pr <- return $! DS.annotateAllFree pr
 putStrLn (show (map fst (Map.toList (DS.free pr))))




 putStrLn ""
 putStrLn "cps-converted: "
 (pr,st) <- return $! St.runState (DS.continuated pr) st
 putStrLn (show pr)

 writeFile "debug/cps.out.scm" $
  "(call-with-current-continuation (lambda (halt) \n" ++ (show pr) ++ "))\n"

 putStrLn ""
 putStrLn "relabeled & free-analyzed: "
 (pr,st) <- return $! St.runState (DS.relabel pr) st
 putStrLn "...relabeled!"
 pr <- return $! DS.annotateAllFree pr
 putStrLn "free variable analysis:"
 putStrLn (show (map fst (Map.toList (DS.free pr))))
 putStrLn ""


 putStrLn "statistics: "
 putStrLn $ "user lambdas: " ++ (show (DS.countUserLambdas pr))
 putStrLn $ "cont lambdas: " ++ (show (DS.countContLambdas pr))


 putStrLn ""
 putStrLn "analyzed: "
 results <- return $! GCFA.run pr

-- vis <- return $! map snd (Map.toList (GCFA.anVisited results))
 -- putStrLn (Might.showWithSep "\n\n" (Map.toList (GCFA.anVisited results)))

 putStrLn $ "# states: " ++ (show (GCFA.anNumVisited results))

  -- Escape analysis.
 let escaping = (GCFA.anEscaping results)
 putStrLn $ "escaping: " ++ (show escaping)
 putStrLn $ "# escaping: " ++ (show (Set.size escaping))
 let escapingConts = Set.filter DS.isContTerm escaping
 let escapingUsers = Set.filter DS.isUserTerm escaping
 putStrLn $ "# conts escaping: " ++ (show (Set.size escapingConts))
 putStrLn $ "# users escaping: " ++ (show (Set.size escapingUsers))


  -- Dead code analysis.
 let dead = (GCFA.anDead results)
 putStrLn $ "dead call sites:"
 putStrLn $ Might.showWithSep "\n\n" (Set.toList dead)

 pr <- return $! DS.semDeadRemove dead pr
 writeFile "debug/no-sem-dead.out.scm" $ 
  "(call-with-current-continuation (lambda (halt) \n" ++ (show pr) ++ "))\n"


 
--  putStrLn $ (show (GCFA.singularValues (GCFA.anVisited results)))



 return $! ()

