module Compiler.Backend where

 -- TEMPORARY BUGS:

 -- If you supply a keyword argument, the keyword stays "set" to the
 -- value.  In an ill-formed program, another callee may accidentally
 -- use this keyword.  Callees need to clear ALL the keyword arguments
 -- supplied by the callers.  This means caller's have to inform
 -- callee's which keyword registers are hot.

 -- There is no garbage collection yet.

 -- NOTES:

 -- This C-based backend is GCC-specific, making use of first-class
 -- labels for goto. We should probably compile HOIL to an RTL and
 -- then to C.

import qualified Compiler.Symbol as Sym
import qualified Compiler.HOIL as DS

import qualified Data.Map as Map
import qualified Data.Set as Set

cMangle [] = []
cMangle ('_':tl) = '_':'_':(cMangle tl)
cMangle ('-':tl) = "_dash_" ++ (cMangle tl)
cMangle (c:tl) = c:(cMangle tl)

class Show a => ShowC a where 
 showC :: a -> String

 showC a = cMangle (show a)

instance ShowC (DS.DSExp ann) 
instance ShowC (Sym.Symbol)
 

 -- Receive arguments at the start of a function.
getGCCArgs i (v:tl) = 
 " " ++ (showC v) ++ " = " ++ "$_a" ++ (show i) ++ "_;\n" ++
 (getGCCArgs (i+1) tl)
getGCCArgs _ [] = ""

getGCCKeywords [] = ""
getGCCKeywords ((s,v,a):tl) =
 " " ++ (showC v) ++ " = " ++ "$_key_" ++ (showC s) ++ ";\n" ++
 (getGCCKeywords tl)



 -- Pass arguments before function invocation.
setGCCArgs i (v:tl) = 
 " " ++ "$_a" ++ (show i) ++ "_ = " ++ (trivExpToGCC v) ++ ";\n" ++
 (setGCCArgs (i+1) tl)
setGCCArgs _ [] = ""


setGCCKeywords [] = ""
setGCCKeywords ((s,e,a):tl) = 
 " " ++ "$_key_" ++ (showC s) ++ " = " ++ (trivExpToGCC e) ++ ";\n" ++
 (setGCCKeywords tl)


setEnvGCC env =
 "$_env_ = " ++ (trivExpToGCC env) ++ ";"




bindCountToGCCDecs bindCount = 
 let
  varCount = Map.toAscList bindCount

  generateDecs [] = ""
  generateDecs ((v,n):tl) =
   "cpm_obj_t " ++ (showC v) ++ ";\n" ++ (generateDecs tl)

 in generateDecs varCount


keywordsToGCCDecs keySet = 
 let
  generateDecs [] = ""
  generateDecs (hd:tl) =
   "cpm_obj_t " ++ "$_key_" ++ (showC hd) ++ ";\n" ++ (generateDecs tl)
 in generateDecs (Set.toList keySet)


mainFunctionGCC = 
 "\nvoid prog();\n\n" ++
 "int main(int argc, char* argv[]) {\n" ++
 " init_nova();\n" ++
 " prog();\n" ++
 " return 0;\n" ++
 "}\n\n"



registerDecsGCC = 
 "cpm_obj_t $_env_ ;\n" ++
 "cpm_obj_t $_hnd_ ;\n" ++
 "cpm_obj_t $_cc_ ;\n" ++
 "cpm_obj_t $_aN_ ;\n" ++
 "cpm_obj_t $_a0_ ;\n" ++
 "cpm_obj_t $_a1_ ;\n" ++
 "cpm_obj_t $_a2_ ;\n" ++
 "cpm_obj_t $_a3_ ;\n" ++
 "cpm_obj_t $_a4_ ;\n" ++
 "cpm_obj_t $_a5_ ;\n" ++
 "cpm_obj_t $_a6_ ;\n" ++
 "cpm_obj_t $_a7_ ;\n" ++
 "cpm_obj_t $_a8_ ;\n" ++
 "cpm_obj_t $_ar_ ;\n"


cpmToGCC pr@(DS.LetRec varexps body annt) =
 let
  keywords = DS.findKeywords pr
  bindCount = DS.countBindings pr
 in
    "#include \"cpm_preamble.h\"" ++
    (mainFunctionGCC) ++
    "void prog() {\n\n" ++
    registerDecsGCC ++ 
    (bindCountToGCCDecs bindCount) ++ "\n" ++
    (keywordsToGCCDecs keywords) ++ "\n" ++
    "_main_:\n" ++
    (callToGCC body) ++ "\n" ++
    (varExpsToGCCDefs varexps) ++ "\n" ++
    "}\n"
    -- TODO: body




trivExpToGCC (DS.Nil{}) = "CPM_NIL";
trivExpToGCC (DS.Bool True annt) = "CPM_TRUE"
trivExpToGCC (DS.Bool False annt) = "CPM_FALSE"
trivExpToGCC (DS.Int i annt) = "CPM_INT32(" ++ (show i)  ++ ")"
trivExpToGCC (DS.Var sym annt) = 
 case DS.boundVars annt of
  Just bmap -> 
   case Map.lookup sym bmap of 
    Just (DS.VarInfo depth DS.LetRecBound) -> 
      "CPM_FUNLAB(&&" ++ (showC sym) ++ ")"
    Just (DS.VarInfo depth _) -> (showC sym)
trivExpToGCC (DS.Symbol sym annt) =
 "CPM_SYMBOL(" ++ (show $ Sym.toCode sym) ++ ")"

trivExpToGCC (DS.Prim{DS.primop=DS.Halt{}}) = "CPM_HALT"

trivExpToGCC (DS.App (DS.Prim{DS.primop=
			      DS.Halt{}})
              []
	      (val:tl) annt) =
 "cpm_halt(" ++ (trivExpToGCC val) ++ ")"


trivExpToGCC (DS.App (DS.Prim{DS.primop=
			      DS.Difference{}})
              []
	      [a,b] annt) =
 "cpm_difference(" ++ (trivExpToGCC a) ++ ", " ++ (trivExpToGCC b) ++ ")"

trivExpToGCC (DS.App (DS.Prim{DS.primop=
			      DS.Product{}})
	      []
	      [a,b] annt) =
 "cpm_product(" ++ (trivExpToGCC a) ++ ", " ++ (trivExpToGCC b) ++ ")"

trivExpToGCC (DS.App (DS.Prim{DS.primop=
			      DS.Sum{}})
	      []
	      [a,b] annt) =
 "cpm_sum(" ++ (trivExpToGCC a) ++ ", " ++ (trivExpToGCC b) ++ ")"

trivExpToGCC (DS.App (DS.Prim{DS.primop=
			      DS.Select{DS.selectIndex=i}}) 
                        []
                        [exp] annt) = 
 "cpm_select(" ++ (trivExpToGCC exp) ++ ", " ++ (show i) ++ ")"

trivExpToGCC (DS.App (DS.Prim{DS.primop=DS.Tuple{}}) [] [] annt) =
 "CPM_EMPTYTUPLE"

trivExpToGCC (DS.App (DS.Prim{DS.primop=DS.Tuple{}}) [] exps annt) =
 "cpm_tuple(" ++ (show $ length exps) ++ ", " ++ (trivExpsToGCC exps) ++ ")"



 -- Relational operators.

trivExpToGCC (DS.App (DS.Prim{DS.primop=DS.EqNum{}}) [] [a,b] annt) =
 "cpm_eqnum(" ++ (trivExpToGCC a) ++ ", " ++ (trivExpToGCC b) ++ ")"

trivExpToGCC (DS.App (DS.Prim{DS.primop=DS.LessThanEq{}}) [] [a,b] annt) =
 "cpm_lteq(" ++ (trivExpToGCC a) ++ ", " ++ (trivExpToGCC b) ++ ")"


 -- Side effects.
trivExpToGCC (DS.App (DS.Prim{DS.primop=DS.CellNew{}}) [] [exp] annt) =
 "cpm_cellnew(" ++ (trivExpToGCC exp) ++ ")"

trivExpToGCC (DS.App (DS.Prim{DS.primop=DS.CellContents{}}) [] [loc] annt) =
 "cpm_cellcontents(" ++ (trivExpToGCC loc) ++ ")"

trivExpToGCC (DS.App (DS.Prim{DS.primop=DS.CellSet{}}) [] [loc,val] annt) =
 "cpm_cellset(" ++ (trivExpToGCC loc) ++ ", " ++ (trivExpToGCC val) ++ ")"



trivExpToGCC unhandled = error $ "unhandled in trivExpToGCC: " ++ (show unhandled)

trivExpsToGCC [] = ""
trivExpsToGCC [hd] = (trivExpToGCC hd) 
trivExpsToGCC (hd:tl) = (trivExpToGCC hd) ++ ", " ++ (trivExpsToGCC tl)



callToGCC (DS.Let (v,e,a) body annt) = 
 " " ++ (showC v) ++ " = " ++ (trivExpToGCC e) ++ ";\n" ++
 (callToGCC body)

callToGCC (DS.If cond cons alt DS.Direct annt) = 
 " if (cpm_istrue(" ++ (trivExpToGCC cond) ++ ")) {\n"  ++
  (callToGCC cons) ++
 " } else {\n" ++
  (callToGCC alt) ++
 " }\n"
 

callToGCC (DS.If cond cons alt DS.ContinuationPassing annt) =
 " if (cpm_istrue(" ++ (trivExpToGCC cond) ++ ")) {\n"  ++
  (callToGCC (DS.sugAppClo cons [] [] DS.nullAnn)) ++
 " } else {\n" ++
  (callToGCC (DS.sugAppClo alt [] [] DS.nullAnn)) ++
 " }\n"

callToGCC (DS.Seq first second annt) =
 if DS.isTrivExp first 
 then " " ++ (trivExpToGCC first) ++ ";\n " ++ (callToGCC second) ++ ";\n" 
 else error $ "Non-trivial expression in Seq." 

callToGCC (DS.App (DS.Prim{DS.primop=DS.Halt{}}) [] [retval] annt) =
 " cpm_halt(" ++ (trivExpToGCC retval) ++ ");\n"



callToGCC (DS.App f keywords args annt) =
 (setGCCKeywords keywords) ++ 
 (setGCCArgs 0 args) ++
 " CPM_CALL(" ++ (trivExpToGCC f) ++ ");\n" 





callToGCC unhandled = 
 error $ "unhandled in callToGCC: " ++ (show unhandled)


varExpsToGCCDefs [] = ""
varExpsToGCCDefs ((v,DS.Lam keywords formals body an,a):tl) =
 (showC v) ++ ":\n" ++
-- ((show env) ++ " = $_env_ ;") ++
 (getGCCKeywords keywords) ++ 
 (getGCCArgs 0 formals) ++ 
 (callToGCC body) ++
 (varExpsToGCCDefs tl) ++
 "\n"
