module Compiler.ChickenBackend where

 -- TODO/SOFTWARE ENGINEERING: qualify this import:
import Compiler.HOIL as DS -- Direct Style IR.

import qualified Might

chickenizedPrimOp (EqNum checks hnd) = 
  "(prim-= " ++ 
   "'ds" ++
   " '(" ++ (Might.showWithSep " " checks) ++ ") " ++
   (showExn hnd) ++ 
  ")"

chickenizedPrimOp (LessThan checks hnd) =
  "(prim-< " ++ 
   "'ds" ++
   " '(" ++ (Might.showWithSep " " checks) ++ ") " ++
   (showExn hnd) ++ 
  ")"

chickenizedPrimOp (LessThanEq checks hnd) = 
  "(prim-<= " ++ 
   "'ds" ++
   " '(" ++ (Might.showWithSep " " checks) ++ ") " ++
   (showExn hnd) ++ 
  ")"

chickenizedPrimOp (Sum checks hnd) =
  "(prim-+ " ++ 
   "'ds" ++
   " '(" ++ (Might.showWithSep " " checks) ++ ") " ++
   (showExn hnd) ++ 
  ")"

chickenizedPrimOp (Difference checks hnd) =
  "(prim-- " ++ 
   "'ds" ++
   " '(" ++ (Might.showWithSep " " checks) ++ ") " ++
   (showExn hnd) ++ 
  ")"  

chickenizedPrimOp (Product checks hnd) = 
  "(prim-* " ++ 
   "'ds" ++
   " '(" ++ (Might.showWithSep " " checks) ++ ") " ++
   (showExn hnd) ++ 
  ")"  


chickenizedPrimOp (Tuple hnd) = 
  "(prim-tuple " ++ 
   "'ds" ++
   " " ++ 
   (showExn hnd) ++ 
  ")"  


chickenizedPrimOp (Select index checks hnd) = 
  "(prim-select " ++ 
   (show index) ++
   " " ++
   "'ds" ++
   " '(" ++ (Might.showWithSep " " checks) ++ ") " ++
   (showExn hnd) ++ 
  ")"  

chickenizedPrimOp (Halt{}) = "halt"


  -- TODO: Properly print these primops.
chickenizedPrimOp (Cons{}) = error "cons"
chickenizedPrimOp (Car{}) = error "car"
chickenizedPrimOp (Cdr{}) = error "cdr" 
chickenizedPrimOp (CellNew{}) = "cell-new"
chickenizedPrimOp (CellSet{}) = "cell-set!"
chickenizedPrimOp (CellContents{}) = "cell-contents"



chickenizedVarExps [] = ""
chickenizedVarExps [(v,e,a)] = "(" ++ (show v) ++ " " ++ (chickenized e) ++ ")"
chickenizedVarExps ((v,e,a):tl) = (chickenizedVarExps [(v,e,a)]) ++ "\n" ++ (chickenizedVarExps tl)


chickenizedKeywords [] = ""
chickenizedKeywords [(s,e,a)] = (show s) ++ ": " ++ (chickenized e)
chickenizedKeywords (hd:tl) =
 (chickenizedKeywords [hd]) ++ " " ++ (chickenizedKeywords tl)

chickenizedKeywordsWithBody [] body
 = chickenized body
chickenizedKeywordsWithBody ((s,v,a):tl) body
 = " (let ((" ++ (show v) ++ " " 
 ++ "(get-keyword #:" ++ (show s) ++ " chicken-key-args)))\n" 
 ++ (chickenizedKeywordsWithBody tl body) ++ ")"
 

chickenizedWithSep :: String -> [DSExp ext] -> String
chickenizedWithSep sep [] = ""
chickenizedWithSep sep [last]
 = chickenized last
chickenizedWithSep sep (hd:tl)
 = (chickenized hd) ++ sep ++ (chickenizedWithSep sep tl)



chickenized (Nil{}) = "'()"
chickenized (Bool{bool=True}) = "#t"
chickenized (Bool{bool=False}) = "#f"
chickenized (Int{int=i}) = show i
chickenized (Var{symbol=s}) = show s
chickenized (Symbol{symbol=s}) = "'" ++ (show s)
chickenized (Set{varexp=(v,e,a)}) =
 "(set! " ++ (chickenized v) ++ " " ++ (chickenized e) ++ ")\n"
chickenized (Seq{firstExp=e1,secondExp=e2}) =
 "(begin " ++ (chickenized e1) ++ "\n " ++ (chickenized e2) ++ ")"
chickenized (Lam [] formals body annt) = 
 "(lambda (" ++ (Might.showWithSep " " formals) ++ ")\n "
   ++ (chickenized body) ++ ")"
chickenized (Lam keywords formals body annt) = 
 "(lambda (" ++ (Might.showWithSep " " formals) ++ " . chicken-key-args)\n"
   ++ " " ++ (chickenizedKeywordsWithBody keywords body) ++ ")\n "
chickenized (Let (v,e,a) body ann) = 
 "(let ((" ++ (chickenized v) ++ " " ++ (chickenized e) ++ "))\n " ++ (chickenized body) ++ ")"

chickenized (LetRec varexps body ann) =
 "(letrec (" ++ (chickenizedVarExps varexps) ++ ")\n " ++ (chickenized body) ++ ")"
chickenized (If cond cons alt Direct ann) =
 "(if " ++ (chickenized cond) ++ " " ++ (chickenized cons) ++ " " ++ (chickenized alt) ++ ")"
chickenized (If cond cons alt ContinuationPassing ann) =
 "(if-cps " ++ (chickenized cond) ++ " " ++ (chickenized cons) ++ " " ++ (chickenized alt) ++ ")"
chickenized (App f [] [] ann) = 
 "(" ++ (chickenized f) ++ ")"
chickenized (App f [] args ann) = 
 "(" ++ (chickenized f) ++ " " ++ (chickenizedWithSep " " args) ++ ")"
chickenized (App f keywords args ann) = 
 "(" ++ (chickenized f) ++ " " ++ (chickenizedWithSep " " args) 
     ++ " " ++ (chickenizedKeywords keywords)
     ++ ")"

chickenized (Prim{primop=p}) = (chickenizedPrimOp p)
chickenized (Try exp handler ann) =
 "(try " ++ (chickenized exp) ++ " " ++ (chickenized handler) ++ ")"
chickenized (Raise exn ann) =
 "(raise " ++ (chickenized exn) ++ ")"
chickenized (CallCC False ann) =
 "call/cc" 
chickenized (CallCC True ann) = 
 "call/cc+exn"
