 -- A library for generic S-Expressions.
 -- WARNING, BUG, TODO: Incomplete.
module Compiler.SExpression where

-- import Monad (liftM)
import Text.ParserCombinators.Parsec hiding (spaces)
import Control.Monad.State as St

import qualified Monad (liftM)

import qualified Compiler.Symbol as Symbol


data SExp annotation external =
   -- Pre-canonicalized S-Expressions.

    -- Used only on first parse.
   RawSymbol { rawSymbol :: !String , 
	       ann :: annotation }


   -- Syntactic S-Expressions.
 | Nil { ann :: annotation }

 | Cons { car :: !(SExp annotation external),
          cdr :: !(SExp annotation external),
	  ann :: annotation }

 | Symbol { symbol :: !Symbol.Symbol,
	    ann :: annotation }

 | Integer { integer :: !Integer ,
	     ann :: annotation }

 | Float { real :: !Float ,
	   ann :: annotation }

 | Int { int :: !Int ,
	 ann :: annotation }

 | Char { character :: !Char ,
	  ann :: annotation }

 | Bool { bool :: !Bool,
	  ann :: annotation }

 | String { string :: !String ,
	    ann :: annotation }

 | Vector { elements :: ![SExp annotation external],
            ann :: annotation }

   -- Allows for Semantic S-Expressions.
 | Ext { ext :: external , 
	 ann :: annotation }



instance Show (SExp ann ext) where 
 show (Nil{}) = "()"
 show (Bool{bool=True}) = "#t"
 show (Bool{bool=False}) = "#f"
 show (Char{character=c}) = "#\\" ++ [c]
 show (Symbol{symbol=s}) = (show s)
 show (Int{int=i}) = (show i)
 show (String s ann) = "\"" ++ s ++ "\""
 show sx@(Cons hd tl ann) = 
  showCons sx 
   where 
    showCons (Cons{car=hd,cdr=tl}) =
     "(" ++ (show hd) ++ (showConsTail tl) 
    showCons (Nil{}) = "()"

    showConsTail (Cons{car=hd,cdr=tl}) =
     " " ++ (show hd) ++ (showConsTail tl)
    showConsTail (Nil{}) = ")"
    showConsTail (oth) = " . " ++ (show oth) ++ ")"




kar (Cons l r a) = l
kar other = error $ "kar of: " ++ (show other)

kdr (Cons l r a) = r

isCons (Cons l r a) = True
isCons _ = False


 -- Source allows lexed S-Expressions to be annotated.
data Source = 
   SourceUndefined 
 | SourceFile !String !Int !Int


 -- Lexical S-Expressions come directly from input.
type LexicalSExp = SExp Source ()



 -- Symbol canonicalization uniquifies symbols, if desired.


canonicalizeMany :: Symbol.HasSymbolTable a => 
                 [SExp annt ext] -> St.State a [SExp annt ext]
canonicalizeMany [] =
 return $! []
canonicalizeMany (hd:tl) =
 do
 hd <- canonicalize hd
 tl <- canonicalizeMany tl
 return $! hd:tl  



canonicalize :: Symbol.HasSymbolTable a => 
                 (SExp annt ext) -> St.State a (SExp annt ext)
canonicalize (RawSymbol string ann) =
 do sym <- Symbol.symbolize string
    return $! Symbol sym ann

canonicalize (Cons car cdr ann) =
 do 
    car <- canonicalize car
    cdr <- canonicalize cdr
    return $! Cons car cdr ann

canonicalize atom = return $! atom




 -- Simple injectors.
cons a b = Cons a b SourceUndefined

fromListWithTail tail [] = tail
fromListWithTail tail (hd:tl) = cons hd (fromListWithTail tail tl)

fromList = fromListWithTail (Nil SourceUndefined)

fromSymbol sym = Symbol sym SourceUndefined

fromInt int = Int int SourceUndefined

nil = Nil SourceUndefined


symbolFromString :: String -> LexicalSExp
symbolFromString string = RawSymbol string SourceUndefined


functor :: String -> LexicalSExp -> LexicalSExp
functor string body = fromList[symbolFromString string,body]


mapToList f sexp = map f (toList sexp)

mapToImproperList f sexp = (map f lst, Monad.liftM f tail)
 where (lst,tail) = toImproperList sexp



 -- Extractors.
toList (Cons car cdr _) = car:(toList cdr)
toList (Nil{}) = []
toList sx = error $ "Improper list in list's place: " ++ (show sx)


toImproperList (Cons car cdr _) = 
 let 
  (cdr',tail) = toImproperList cdr
 in (car:cdr',tail)
toImproperList (Nil{}) = ([],Nothing)
toImproperList sx = ([],Just sx)

 -- Predicates.
isSymbol (Symbol{}) = True
isSymbol _ = False


 -- Lexing & parsing.
sexp_symbol :: Parser Char
sexp_symbol = oneOf "!$#%&|*+-/<=>?@^_~"

comment :: Parser [Char]
comment = 
  do c <- char ';' 
     rest <- manyTill anyChar (try newline)
     return $! c:rest

whitespaces =
 do skipMany (realspaces <|> comment)

realspaces :: Parser [Char]
realspaces = many1 space


getFileSource :: Parser Source
getFileSource = 
 do pos <- getPosition
    return $! SourceFile 
	       (sourceName pos) 
	       (sourceLine pos)
	       (sourceColumn pos)


parseString :: Parser LexicalSExp
parseString = do 
                 char '"'
		 pos <- getFileSource
                 x <- many (noneOf "\"")
                 char '"'
                 return $! String x pos


parseAtom :: Parser LexicalSExp
parseAtom = do
	       pos <- getFileSource
               first <- char '.' <|> letter <|> sexp_symbol
               rest <- many (letter <|> digit <|> sexp_symbol)
               let atom = [first] ++ rest
               return $! case atom of 
                           "#t" -> Bool True pos
                           "#f" -> Bool False pos
                           otherwise -> RawSymbol atom pos


parseNumber :: Parser LexicalSExp
parseNumber = 
 do
    pos <- getFileSource
    n <- many1 digit
    return $! (Int (read n) pos)


parseCompExpr :: Parser LexicalSExp
parseCompExpr = 
 do
 sx <- parseExpr 
 -- sx <- parseExprTail sx
 return $! sx
 


parseExpr :: Parser LexicalSExp
parseExpr =
 do
 try whitespaces
 sx <- (parseAtom
  <|> parseString
  <|> parseNumber
  <|> parseQuoted
  <|> do
        char '('
        x <- (try parseList) <|> parseDottedList
        char ')'
        return $! x
  <|> do 
        pos <- getFileSource
        char '[' 
        x <- (try parseList) <|> parseDottedList
        char ']'
        return $! Cons (Symbol Symbol.block_quote' pos) x pos
  <|> do 
        pos <- getFileSource
        char '#' 
        char '(' 
        x <- (try parseList) <|> parseDottedList
        char ')'
        return $! Cons (Symbol Symbol.vector' pos) x pos)
                        
 return $! sx



parseExprTail head =
 do
 try whitespaces
 more <- option Nothing (do { c <- char ':' <|> char '.' ; return $! Just c ; })
 case more of
  Just ':' ->
   do
   sx <- parseExpr
    -- TODO: Properly report source info.
   parseExprTail (fromList [Symbol Symbol.colon' SourceUndefined,head,sx])
  Just oth ->
   error $ "Encountered . while parsing tail of " ++ (show head)
--   Just '.' ->
--    do
--    tail <- parseExpr
--    return $! (Cons head tail SourceUndefined)
  Nothing -> return $! head
 

parseList :: Parser LexicalSExp
parseList = 
 do 
    pos <- getFileSource
    exps <- sepBy parseCompExpr whitespaces
    return $! (fromListWithTail (Nil SourceUndefined) exps)
	      {ann=pos}


parseDottedList :: Parser LexicalSExp
parseDottedList = do
    pos <- getFileSource
    exps <- endBy parseCompExpr whitespaces
    char '.'
    whitespaces
    tail <- parseCompExpr
    whitespaces
    return $! (fromListWithTail tail exps) {ann=pos}


parseQuoted :: Parser LexicalSExp
parseQuoted = do
    pos <- getFileSource
    char '\''
    x <- parseCompExpr
    return $! (fromList [(symbolFromString "quote"){ ann=pos },x])
                {ann=pos}

parseMany1CompExpr = many1 parseCompExpr


fromStringWithName :: String -> String -> LexicalSExp
fromStringWithName filename input = case parse parseCompExpr filename input of
    Left err -> error $ "No match: " ++ show err
    Right val -> val


listFromStringWithName :: String -> String -> [LexicalSExp]
listFromStringWithName filename input = 
  case parse parseMany1CompExpr filename input of
   Left err -> error $ "No match: " ++ show err
   Right val -> val

