module Lattice.StandardLattice where

import qualified Data.Map as Map
import qualified Data.Set as Set
-- import qualified Math.InfiniteMap as IM
import Lattice.Lattice



--  -- K -> A : A map lattice.
-- instance (Ord k, Lattice a) => Lattice (IM.InfiniteMap k a) where
--  bot = IM.constant bot
--  top = IM.constant top
--  f <: g     = IM.isSubmapOfBy (<:) f g
--  f `join` g = IM.lift join f g
--  f `meet` g = IM.lift meet f g


instance (Ord k, Lattice a) => Lattice (Map.Map k a) where
 bot = Map.empty
 top = error $ "Cannot be represented."
 f <: g = Map.isSubmapOfBy (<:) f g
 f `join` g = Map.unionWith join f g
 f `meet` g = Map.intersectionWith meet f g


at :: (Ord k, Lattice a) => Map.Map k a -> k -> a
f `at` x = case Map.lookup x f of
             Just y -> y
             Nothing -> bot


 -- Maybe A : An option lattice.
instance (Lattice a) => Lattice (Maybe a) where
 top = error $ "there is no top for a Maybe lattice."
 bot = error $ "there is not bot for a Maybe lattice." 

 (Just a) <: (Just b) = a <: b
 Nothing <: Nothing = True
 _  <: _ = False

 (Just a) `join` (Just b) = Just $! a `join` b
 Nothing `join` Nothing = Nothing

 meet a b = error $ "meet undefined on maybe"
 


 -- A x B : A cartesian product lattice.
instance (Lattice a, Lattice b) => Lattice (a,b) where
 bot = (bot,bot)
 top = (top,top)
 (a1,b1) <: (a2,b2) = (a1 <: a2) ||
                      (a1 == a2 && b1 <: b2) 
 (a1,b1) `join` (a2,b2) = (a1 `join` a2, b1 `join` b2)
 (a1,b1) `meet` (a2,b2) = (a1 `meet` a2, b1 `meet` b2) 




 -- A* : A sequence lattice.
instance (Lattice a) => Lattice [a] where
 top = error "top sequence"
 bot = []
 [] <: [] = True
 (hd1:tl1) <: (hd2:tl2) = hd1 <: hd2 && tl1 <: tl2
 [] <: (hd:tl) = False
 (hd:tl) <: [] = False

 [] `join` [] = []
 (hd1:tl1) `join` (hd2:tl2) = (hd1 `join` hd2):(tl1 `join` tl2)

 (hd1:tl1) `meet` (hd2:tl2) = (hd1 `meet` hd2):(tl1 `meet` tl2)

                          

 -- P(L) : A power lattice.
allIn :: Set.Set a -> (a -> Bool) -> Bool
allIn set p = Set.fold (\ x soFar -> soFar && p x) True set

existsIn :: Set.Set a -> (a -> Bool) -> Bool
existsIn set p = Set.fold (\ x soFar -> soFar || p x) False set

instance (Ord a, Lattice a) => Lattice (Set.Set a) where
 bot = Set.empty
 top = error "top set"
 setA <: setB = allIn setA (\a -> 
                 existsIn setB (\b ->
                   a <: b))
 setA `join` setB = Set.union setA setB
 setA `meet` setB = Set.fromList [ x | x <- (Set.toList setA) ++ 
                                            (Set.toList setB),
                                       (Set.singleton x) <: setA,
                                       (Set.singleton x) <: setB ] 



 -- A multitude of cross-product lattices.


instance (Lattice a, Lattice b, Lattice c, Lattice d)
  => Lattice (a,b,c,d) where
 (a1,b1,c1,d1) <: (a2,b2,c2,d2) = 
  a1 <: a2 && b1 <: b2 && c1 <: c2 && d1 <: d2

 (a1,b1,c1,d1) `join` (a2,b2,c2,d2) =
  (a1 `join` a2, b1 `join` b2, c1 `join` c2, d1 `join` d2)

 (a1,b1,c1,d1) `meet` (a2,b2,c2,d2) =
  (a1 `meet` a2, b1 `meet` b2, c1 `meet` c2, d1 `meet` d2)



instance (Lattice a, Lattice b, Lattice c, Lattice d, Lattice e) 
  => Lattice (a,b,c,d,e) where
 (a1,b1,c1,d1,e1) <: (a2,b2,c2,d2,e2) = 
  a1 <: a2 && b1 <: b2 && c1 <: c2 && d1 <: d2 && e1 <: e2 

 (a1,b1,c1,d1,e1) `join` (a2,b2,c2,d2,e2) =
  (a1 `join` a2, b1 `join` b2, c1 `join` c2, d1 `join` d2, e1 `join` e2)

 (a1,b1,c1,d1,e1) `meet` (a2,b2,c2,d2,e2) =
  (a1 `meet` a2, b1 `meet` b2, c1 `meet` c2, d1 `meet` d2, e1 `meet` e2)


instance (Lattice a, Lattice b, Lattice c, Lattice d, Lattice e, Lattice f) 
  => Lattice (a,b,c,d,e,f) where
 (a1,b1,c1,d1,e1,f1) <: (a2,b2,c2,d2,e2,f2) = 
  a1 <: a2 && b1 <: b2 && c1 <: c2 && d1 <: d2 && e1 <: e2 && f1 <: f2 

 (a1,b1,c1,d1,e1,f1) `join` (a2,b2,c2,d2,e2,f2) =
  (a1 `join` a2,
   b1 `join` b2,
   c1 `join` c2,
   d1 `join` d2,
   e1 `join` e2,
   f1 `join` f2)
 (a1,b1,c1,d1,e1,f1) `meet` (a2,b2,c2,d2,e2,f2) =
  (a1 `meet` a2,
   b1 `meet` b2,
   c1 `meet` c2,
   d1 `meet` d2,
   e1 `meet` e2,
   f1 `meet` f2)


instance (Lattice a, Lattice b, Lattice c,  Lattice d, Lattice e, 
          Lattice f, Lattice g) => Lattice (a,b,c,d,e,f,g) where
 (a1,b1,c1,d1,e1,f1,g1) <: (a2,b2,c2,d2,e2,f2,g2) = 
  a1 <: a2 &&
  b1 <: b2 &&
  c1 <: c2 &&
  d1 <: d2 &&
  e1 <: e2 &&
  f1 <: f2 &&
  g1 <: g2 

 (a1,b1,c1,d1,e1,f1,g1) `join` (a2,b2,c2,d2,e2,f2,g2) =
  (a1 `join` a2,
   b1 `join` b2,
   c1 `join` c2,
   d1 `join` d2,
   e1 `join` e2,
   f1 `join` f2,
   g1 `join` g2)

 (a1,b1,c1,d1,e1,f1,g1) `meet` (a2,b2,c2,d2,e2,f2,g2) =
  (a1 `meet` a2,
   b1 `meet` b2,
   c1 `meet` c2,
   d1 `meet` d2,
   e1 `meet` e2,
   f1 `meet` f2,
   g1 `meet` g2)


