module Parser.CParser where

import qualified Control.Monad.State as St
import qualified Char (isSpace,isAlpha,isDigit)

import qualified Data.Set as Set 

data ParserState = 
 PS { filename :: !String ,
      col :: !Int ,
      line :: !Int ,
      typedefs :: ![Set.Set String] ,
      input :: !String }

type ParserMonad = St.State ParserState


getPosition :: St.State ParserState Pos 
getPosition = 
 do
 pstate <- St.get
 return $! Pos (filename pstate) (line pstate) (col pstate)


isTypedefName :: String -> ParserMonad Bool
isTypedefName str = 
 do
 pstate <- St.get
 tydefs <- return $! typedefs pstate
 return $!
  any (\ set -> str `Set.member` set) tydefs

processIdent :: String -> ParserMonad Token
processIdent str = 
 do
 isTyDef <- isTypedefName str
 return $!
  if isTyDef 
  then TokenIdentifier str
  else TokenTypedefName str

processKeyword cont tok = 
 do
 advanceCol $! length (show tok)
 cont $! tok


-- (==>) a b = (a,b)
-- keywords = [
--   "auto" ==> TokenAuto ,
--   "register" ==> 
--  ]

-- parser produced by Happy Version 1.15

data HappyAbsSyn t4 t5
	= HappyTerminal Token
	| HappyErrorToken Int
	| HappyAbsSyn4 t4
	| HappyAbsSyn5 t5

action_0 (4) = happyGoto action_2
action_0 _ = happyReduce_1

action_1 _ = happyFail

action_2 (37) = happyAccept
action_2 _ = happyFail

happyReduce_1 = happySpecReduce_0 4 happyReduction_1
happyReduction_1  =  HappyAbsSyn4
		 (error "unfinished"
	)

happyReduce_2 = happyMonadReduce 0 5 happyReduction_2
happyReduction_2 (happyRest)
	 = happyThen ( getPosition
	) (\r -> happyReturn (HappyAbsSyn5 r))

happyNewToken action sts stk
	= lexer(\tk -> 
	let cont i = action i i tk (HappyState action) sts stk in
	case tk of {
	TokenEOF -> action 37 37 (error "reading EOF!") (HappyState action) sts stk;
	TokenAuto -> cont 6;
	TokenRegister -> cont 7;
	TokenStatic -> cont 8;
	TokenExtern -> cont 9;
	TokenTypedef -> cont 10;
	TokenVoid -> cont 11;
	TokenChar -> cont 12;
	TokenShort -> cont 13;
	TokenInt -> cont 14;
	TokenLong -> cont 15;
	TokenFloat -> cont 16;
	TokenDouble -> cont 17;
	TokenSigned -> cont 18;
	TokenUnsigned -> cont 19;
	TokenConst -> cont 20;
	TokenVolatile -> cont 21;
	TokenStruct -> cont 22;
	TokenUnion -> cont 23;
	TokenIdentifier happy_dollar_dollar -> cont 24;
	TokenTypedefName happy_dollar_dollar -> cont 25;
	TokenEq -> cont 26;
	TokenPlus -> cont 27;
	TokenMinus -> cont 28;
	TokenTimes -> cont 29;
	TokenDiv -> cont 30;
	TokenLPar -> cont 31;
	TokenRPar -> cont 32;
	TokenLBrace -> cont 33;
	TokenRBrace -> cont 34;
	TokenLBrack -> cont 35;
	TokenRBrack -> cont 36;
	_ -> happyError'
	})

happyError_ tk = happyError'

happyThen :: () => St.State ParserState a -> (a -> St.State ParserState b) -> St.State ParserState b
happyThen = (>>=)
happyReturn :: () => a -> St.State ParserState a
happyReturn = (return)
happyThen1 = happyThen
happyReturn1 :: () => a -> St.State ParserState a
happyReturn1 = happyReturn
happyError' :: () => St.State ParserState a
happyError' = happyError

cparse = happySomeParser where
  happySomeParser = happyThen (happyParse action_0) (\x -> case x of {HappyAbsSyn4 z -> happyReturn z; _other -> notHappyAtAll })

happySeq = happyDontSeq

happyError :: ParserMonad a 
happyError = error "Parse error"


data Exp  
      = Let String Exp Exp
      | Exp1 Exp1

data Exp1 
      = Plus Exp1 Term 
      | Minus Exp1 Term 
      | Term Term

data Term 
      = Times Term Factor 
      | Div Term Factor 
      | Factor Factor

data Factor 
      = Int Int 
      | Var String 
      | Brack Exp


data Pos
 = Pos { posFilename :: !String, posLine :: !Int, posCol :: !Int }

data Token
      = TokenAuto
      | TokenRegister
      | TokenStatic
      | TokenExtern
      | TokenTypedef

      | TokenVoid
      | TokenChar
      | TokenShort
      | TokenInt
      | TokenLong
      | TokenFloat
      | TokenDouble
      | TokenSigned
      | TokenUnsigned

      | TokenConst
      | TokenVolatile

      | TokenStruct
      | TokenUnion

      | TokenInteger Int

      | TokenIdentifier String
      | TokenTypedefName String

      | TokenEq
      | TokenPlus
      | TokenMinus
      | TokenTimes
      | TokenDiv

      | TokenLPar
      | TokenRPar

      | TokenLBrace
      | TokenRBrace

      | TokenLBrack
      | TokenRBrack

      | TokenEOF

instance Show Token where
 show TokenAuto = "auto"
 show TokenRegister = "register"
 show TokenStatic = "static"
 show TokenExtern = "extern"
 show TokenTypedef = "typedef"

 show TokenVoid = "void"
 show TokenChar = "char"
 show TokenShort = "short"
 show TokenInt = "int"
 show TokenLong = "long"
 show TokenFloat = "float"
 show TokenDouble = "double"
 show TokenSigned = "signed"
 show TokenUnsigned = "unsigned"

 show TokenConst = "const"
 show TokenVolatile = "volatile"




advanceCol :: Int -> ParserMonad ()
advanceCol numchars =
 do
 pstate <- St.get
 St.put $! pstate { col = (col pstate) + 1, input = tail $! input pstate }


advanceLine :: ParserMonad ()
advanceLine =
 do
 pstate <- St.get
 St.put $! pstate { col = 1, line = (line pstate) + 1, 
                    input = tail $! input pstate }


lexer :: (Token -> ParserMonad a) -> ParserMonad a
lexer cont = 
 do
 pstate <- St.get
 let 
   -- End of file.
  lex [] = cont TokenEOF

   -- Ignore whitespace.
  lex (' ':cs) = do { advanceCol 1 ; lexer cont ; }
  lex ('\t':cs) = do { advanceCol 1 ; lexer cont ; }
  lex ('\n':cs) = do { advanceLine ; lexer cont ; }

   -- Keywords.
  lex ('a':'u':'t':'o':cs) =
   processKeyword cont TokenAuto
  lex ('r':'e':'g':'i':'s':'t':'e':'r':cs) =
   processKeyword cont TokenRegister
  lex ('s':'t':'a':'t':'i':'c':cs) =
   processKeyword cont TokenStatic
  lex ('e':'x':'t':'e':'r':'n':cs) =
   processKeyword cont TokenExtern
  lex ('t':'y':'p':'e':'d':'e':'f':cs) = 
   processKeyword cont TokenTypedef

  lex ('v':'o':'i':'d':cs) =
   processKeyword cont TokenVoid
  lex ('c':'h':'a':'r':cs) =
   processKeyword cont TokenChar
  lex ('s':'h':'o':'r':'t':cs) =
   processKeyword cont TokenShort
  lex ('i':'n':'t':cs) =
   processKeyword cont TokenInt
  lex ('l':'o':'n':'g':cs) =
   processKeyword cont TokenLong
  lex ('f':'l':'o':'a':'t':cs) =
   processKeyword cont TokenFloat
  lex ('d':'o':'u':'b':'l':'e':cs) =
   processKeyword cont TokenDouble
  lex ('s':'i':'g':'n':'e':'d':cs) =
   processKeyword cont TokenSigned
  lex ('u':'n':'s':'i':'g':'n':'e':'d':cs) =
   processKeyword cont TokenUnsigned

  lex ('c':'o':'n':'s':'t':cs) =
   processKeyword cont TokenVoid
  lex ('v':'o':'l':'a':'t':'i':'l':'e':cs) =
   processKeyword cont TokenChar

  lex ('s':'t':'r':'u':'c':'t':cs) =
   processKeyword cont TokenStruct
  lex ('u':'n':'i':'o':'n':cs) =
   processKeyword cont TokenUnion


   -- Parens, braces and brackets.
  lex ('(':cs) = do { advanceCol 1 ; cont TokenLPar ; } 
  lex (')':cs) = do { advanceCol 1 ; cont TokenRPar ; } 

  lex ('{':cs) = do { advanceCol 1 ; cont TokenLBrace ; } 
  lex ('}':cs) = do { advanceCol 1 ; cont TokenRBrace ; } 

  lex ('[':cs) = do { advanceCol 1 ; cont TokenLBrack ; } 
  lex (']':cs) = do { advanceCol 1 ; cont TokenRBrack ; } 


  lex (c:cs) 
      | Char.isDigit c = lexNum [c] (cs)
      | Char.isAlpha c = lexAlpha [c] (cs)


   -- Lex an alphabetic sequence.
  lexAlpha revid [] = do { tok <- processIdent (reverse revid) ; cont tok ; }
  lexAlpha revid (c:cs) 
      | Char.isAlpha c = do { advanceCol 1 ; lexAlpha (c:revid) cs ; } 
      | otherwise = do { tok <- processIdent (reverse revid) ; cont tok ; }
  
   -- Lex a number.
  lexNum revnum [] = error "unfinished"

  in lex $! input pstate



-- lexer :: String -> [Token]
-- lexer [] = []
-- lexer (c:cs) 
--       | Char.isSpace c = lexer cs
--       | Char.isAlpha c = lexVar (c:cs)
--       | Char.isDigit c = lexNum (c:cs)
-- lexer ('=':cs) = TokenEq : lexer cs
-- lexer ('+':cs) = TokenPlus : lexer cs
-- lexer ('-':cs) = TokenMinus : lexer cs
-- lexer ('*':cs) = TokenTimes : lexer cs
-- lexer ('/':cs) = TokenDiv : lexer cs
-- lexer ('(':cs) = TokenLPar : lexer cs
-- lexer (')':cs) = TokenRPar : lexer cs

-- lexNum cs = TokenInt (read num) : lexer rest
--       where (num,rest) = span Char.isDigit cs

-- lexVar cs =
--    case span Char.isAlpha cs of
--       ("let",rest) -> TokenLet : lexer rest
--       ("in",rest)  -> TokenIn : lexer rest
--       (var,rest)   -> TokenVar var : lexer rest

-- main = getContents >>= print . calc . lexer
{-# LINE 1 "GenericTemplate.hs" #-}
{-# LINE 1 "<built-in>" #-}
{-# LINE 1 "<command line>" #-}
{-# LINE 1 "GenericTemplate.hs" #-}
-- $Id: CParser.hs,v 1.1.1.1 2006/11/08 23:15:27 matt Exp $

{-# LINE 28 "GenericTemplate.hs" #-}








{-# LINE 49 "GenericTemplate.hs" #-}

{-# LINE 59 "GenericTemplate.hs" #-}

{-# LINE 68 "GenericTemplate.hs" #-}

infixr 9 `HappyStk`
data HappyStk a = HappyStk a (HappyStk a)

-----------------------------------------------------------------------------
-- starting the parse

happyParse start_state = happyNewToken start_state notHappyAtAll notHappyAtAll

-----------------------------------------------------------------------------
-- Accepting the parse

-- If the current token is (1), it means we've just accepted a partial
-- parse (a %partial parser).  We must ignore the saved token on the top of
-- the stack in this case.
happyAccept (1) tk st sts (_ `HappyStk` ans `HappyStk` _) =
	happyReturn1 ans
happyAccept j tk st sts (HappyStk ans _) = 
	 (happyReturn1 ans)

-----------------------------------------------------------------------------
-- Arrays only: do the next action

{-# LINE 155 "GenericTemplate.hs" #-}

-----------------------------------------------------------------------------
-- HappyState data type (not arrays)



newtype HappyState b c = HappyState
        (Int ->                    -- token number
         Int ->                    -- token number (yes, again)
         b ->                           -- token semantic value
         HappyState b c ->              -- current state
         [HappyState b c] ->            -- state stack
         c)



-----------------------------------------------------------------------------
-- Shifting a token

happyShift new_state (1) tk st sts stk@(x `HappyStk` _) =
     let i = (case x of { HappyErrorToken (i) -> i }) in
--     trace "shifting the error token" $
     new_state i i tk (HappyState (new_state)) ((st):(sts)) (stk)

happyShift new_state i tk st sts stk =
     happyNewToken new_state ((st):(sts)) ((HappyTerminal (tk))`HappyStk`stk)

-- happyReduce is specialised for the common cases.

happySpecReduce_0 i fn (1) tk st sts stk
     = happyFail (1) tk st sts stk
happySpecReduce_0 nt fn j tk st@((HappyState (action))) sts stk
     = action nt j tk st ((st):(sts)) (fn `HappyStk` stk)

happySpecReduce_1 i fn (1) tk st sts stk
     = happyFail (1) tk st sts stk
happySpecReduce_1 nt fn j tk _ sts@(((st@(HappyState (action))):(_))) (v1`HappyStk`stk')
     = let r = fn v1 in
       happySeq r (action nt j tk st sts (r `HappyStk` stk'))

happySpecReduce_2 i fn (1) tk st sts stk
     = happyFail (1) tk st sts stk
happySpecReduce_2 nt fn j tk _ ((_):(sts@(((st@(HappyState (action))):(_))))) (v1`HappyStk`v2`HappyStk`stk')
     = let r = fn v1 v2 in
       happySeq r (action nt j tk st sts (r `HappyStk` stk'))

happySpecReduce_3 i fn (1) tk st sts stk
     = happyFail (1) tk st sts stk
happySpecReduce_3 nt fn j tk _ ((_):(((_):(sts@(((st@(HappyState (action))):(_))))))) (v1`HappyStk`v2`HappyStk`v3`HappyStk`stk')
     = let r = fn v1 v2 v3 in
       happySeq r (action nt j tk st sts (r `HappyStk` stk'))

happyReduce k i fn (1) tk st sts stk
     = happyFail (1) tk st sts stk
happyReduce k nt fn j tk st sts stk
     = case happyDrop (k - ((1) :: Int)) sts of
	 sts1@(((st1@(HappyState (action))):(_))) ->
        	let r = fn stk in  -- it doesn't hurt to always seq here...
       		happyDoSeq r (action nt j tk st1 sts1 r)

happyMonadReduce k nt fn (1) tk st sts stk
     = happyFail (1) tk st sts stk
happyMonadReduce k nt fn j tk st sts stk =
        happyThen1 (fn stk) (\r -> action nt j tk st1 sts1 (r `HappyStk` drop_stk))
       where sts1@(((st1@(HappyState (action))):(_))) = happyDrop k ((st):(sts))
             drop_stk = happyDropStk k stk

happyDrop (0) l = l
happyDrop n ((_):(t)) = happyDrop (n - ((1) :: Int)) t

happyDropStk (0) l = l
happyDropStk n (x `HappyStk` xs) = happyDropStk (n - ((1)::Int)) xs

-----------------------------------------------------------------------------
-- Moving to a new state after a reduction

{-# LINE 239 "GenericTemplate.hs" #-}
happyGoto action j tk st = action j j tk (HappyState action)


-----------------------------------------------------------------------------
-- Error recovery ((1) is the error token)

-- parse error if we are in recovery and we fail again
happyFail  (1) tk old_st _ stk =
--	trace "failing" $ 
    	happyError_ tk

{-  We don't need state discarding for our restricted implementation of
    "error".  In fact, it can cause some bogus parses, so I've disabled it
    for now --SDM

-- discard a state
happyFail  (1) tk old_st (((HappyState (action))):(sts)) 
						(saved_tok `HappyStk` _ `HappyStk` stk) =
--	trace ("discarding state, depth " ++ show (length stk))  $
	action (1) (1) tk (HappyState (action)) sts ((saved_tok`HappyStk`stk))
-}

-- Enter error recovery: generate an error token,
--                       save the old token and carry on.
happyFail  i tk (HappyState (action)) sts stk =
--      trace "entering error recovery" $
	action (1) (1) tk (HappyState (action)) sts ( (HappyErrorToken (i)) `HappyStk` stk)

-- Internal happy errors:

notHappyAtAll = error "Internal Happy error\n"

-----------------------------------------------------------------------------
-- Hack to get the typechecker to accept our action functions







-----------------------------------------------------------------------------
-- Seq-ing.  If the --strict flag is given, then Happy emits 
--	happySeq = happyDoSeq
-- otherwise it emits
-- 	happySeq = happyDontSeq

happyDoSeq, happyDontSeq :: a -> b -> b
happyDoSeq   a b = a `seq` b
happyDontSeq a b = b

-----------------------------------------------------------------------------
-- Don't inline any functions from the template.  GHC has a nasty habit
-- of deciding to inline happyGoto everywhere, which increases the size of
-- the generated parser quite a bit.

{-# LINE 303 "GenericTemplate.hs" #-}
{-# NOINLINE happyShift #-}
{-# NOINLINE happySpecReduce_0 #-}
{-# NOINLINE happySpecReduce_1 #-}
{-# NOINLINE happySpecReduce_2 #-}
{-# NOINLINE happySpecReduce_3 #-}
{-# NOINLINE happyReduce #-}
{-# NOINLINE happyMonadReduce #-}
{-# NOINLINE happyGoto #-}
{-# NOINLINE happyFail #-}

-- end of Happy Template.
