;; Standard libary.

;; Tail-recursive reverse.
(define (reverse lst) 

  (define (reverse-help rev lst)  
    (if (null? lst)
        rev
        (reverse-help (cons (car lst) rev) (cdr lst))))
  
  (reverse-help '() lst))


;; Tail-recursive map.
(define (map f lst) 
  
  (define (map-help rev f lst)
    (if (null? lst)
        (reverse rev)
        (map-help (cons (f (car lst)) rev) f (cdr lst))))
  
  (map-help '() f lst))

;; For-each.
(define (for-each f lst)
  (if (null? lst)
      'no-return-value
      (begin
        (f (car lst))
        (for-each f (cdr lst)))))
           
  

;; List helpers.
(define (cadr lst) (car (cdr lst)))
(define (cdar lst) (cdr (car lst)))
(define (cddr lst) (cdr (cdr lst)))
(define (caar lst) (car (car lst)))
(define (caddr lst) (car (cdr (cdr lst))))
(define (cadddr lst) (car (cdr (cdr (cdr lst)))))


;; List membership.
(define (memv x lst)
  (if (null? lst)
      #f
      (if (eqv? (car lst) x)
          lst
          (memv x (cdr lst))))) 


(define (memq x lst)
  (if (null? lst)
      #f
      (if (eq? (car lst) x)
          lst
          (memq x (cdr lst))))) 


(define (member x lst)
  (if (null? lst)
      #f
      (if (equal? (car lst) x)
          lst
          (member x (cdr lst))))) 


;; List length.
(define (length lst) 
 
  ; A local helper.
  (define (length-helper len lst)
    (if (null? lst)
        len
        (length-helper (+ 1 len) (cdr lst))))
  
  (length-helper 0 lst))


;; Associative lists.
(define (assq x f)
  (cond
    ((null? f) #f)
    ((pair? (car f))
     (if (eq? (car (car f)) x)
         (car f)
         (assq x (cdr f))))
    (else ; not a proper associative list, but try anyway.
     (assq x (cdr f)))))


;; Vectors and lists.
(define (vector->list vect)
  (let ((len (vector-length vect)))
    (do ((i    (- len 1) (- i 1))
         (lst '()        (cons (vector-ref vect i) lst)))
      ((< i 0) lst))))

(define (list->vector lst)
  (let* ((len (length lst))
         (vect (make-vector len)))
    (do ((i 0    (+ i 1))
         (l lst  (cdr l)))
      ((null? l) vect)
      (vector-set! vect i (car l)))))
  
         

;; Numerics.
(define (positive? n) (> n 0))
